import numpy as np
import math
import itertools
import scipy as sc
import copy
from threading import Thread
from multiprocessing import Pool

def death(iPop, pdt):
    return iPop[pdt,:]
    
def divide(iPop, pdv):
    return sc.sparse.vstack([iPop, copy.deepcopy(iPop[pdv,:])]).tocsr()
    
def mutate(iPop, pdm, mut_effect):
    cells = len(pdm)
    if cells == 0:
        return iPop
    (a,b) = iPop._shape
    rows = copy.deepcopy(iPop[pdm,:]).toarray()
    rows[:, 0] = rows[:, 0]*(1+abs(mut_effect))**(1 - 2*(mut_effect < 0))
    rows[:, 1] = rows[:, 1] + 1
    iPop = sc.sparse.vstack([iPop, rows]).tocsr()
    (c,d) = iPop._shape
    return iPop

def deleteZeroColumn(iPop):
    mask = np.ones(iPop._shape[1], dtype=bool)
    for i in range(iPop._shape[1]):
        _sum = (iPop[:,i]).count_nonzero()
        if _sum == 0:
            mask[i] = False
    return iPop[:,mask]

def seEvo1Dnorm(iPop, cap, tau_x, A, mut_prob, mut_effect, simTime):
    popSize = iPop._shape[0]
    
    mdt = (popSize/cap)**A
    tau = tau_x * cap/popSize
    mdv = iPop[:,0].toarray()[:,0]
    
    simTime = simTime + tau
    
    pdt = np.random.exponential(1, popSize)/mdt
    pdv = np.random.exponential(1, popSize)
    pdm = np.random.binomial(1, mut_prob, popSize)
    
    pdt = pdt > tau
    pdv = pdv/mdv < tau
    pdm = pdm > np.zeros(popSize)
    
    pdv = pdv & pdt
    pdm = pdm & pdv
    pdv[pdm] = False
    
    nr = np.array(range(popSize))
    pdv = nr[pdv]
    pdt = nr[pdt]
    pdm = nr[pdm]
    
    iPop = divide(iPop, pdv)
    
    iPop = mutate(iPop, pdm, mut_effect) 
    
    pdt = np.append(pdt, range(max(pdt)+1, iPop._shape[0]))
    iPop = death(iPop, pdt)
    
    # iPop = deleteZeroColumn(iPop)
    
    return iPop, simTime


import numpy as np
import math
import itertools
import scipy as sc
import copy
from threading import Thread
from multiprocessing import Pool

def mdv(mut_num, mut_effect):
    if mut_effect > 0:
        return (1+mut_effect)**mut_num
    else:
        return 1/((1-mut_effect)**mut_num)

def mf_dif(iPop, mdt, pdm, mut_effect):
    alfa = 1
    nPop = np.zeros((iPop._shape[0], iPop._shape[1]))
    nPop[:,0] = np.array([x for x in range(int(iPop[0,0]),int(iPop[iPop._shape[0]-1,0])+1)])
    A = - mdt * iPop[:,1].toarray()
    B = (1 - pdm) * mdv(iPop[:,0].toarray(), mut_effect) * iPop[:,1].toarray() * (iPop[:,1].toarray() >= 1)
    C = np.zeros((iPop._shape[0], 1))
    C[1:len(C)] = mdv(iPop[0:iPop._shape[0]-1,0].toarray(), mut_effect) * pdm * iPop[0:iPop._shape[0]-1,1].toarray() * (iPop[0:iPop._shape[0]-1,1].toarray() >= 1)
    nPop = sc.sparse.csr_matrix(nPop)
    nPop[:,1] = A + B + C 
  
    return nPop
    
def rk4(iPop, tau, mdt, pdm, mut_effect):
    k1 = sc.sparse.csr_matrix(np.zeros((iPop._shape[0], iPop._shape[1])))
    k2 = sc.sparse.csr_matrix(np.zeros((iPop._shape[0], iPop._shape[1])))
    k3 = sc.sparse.csr_matrix(np.zeros((iPop._shape[0], iPop._shape[1])))
    k4 = sc.sparse.csr_matrix(np.zeros((iPop._shape[0], iPop._shape[1])))
    k1[:,1] = tau*mf_dif(iPop, mdt, pdm, mut_effect)[:,1]
    k2[:,1] = tau*mf_dif(iPop+k1/2, mdt, pdm, mut_effect)[:,1]
    k3[:,1] = tau*mf_dif(iPop+k2/2, mdt, pdm, mut_effect)[:,1]
    k4[:,1] = tau*mf_dif(iPop+k3, mdt, pdm, mut_effect)[:,1]
    iPop = iPop+(1/6)*(k1+2*k2+2*k3+k4)
    return iPop

def seEvo1Danalytical(iPop, cap, tau_x, A, mut_prob, mut_effect, simTime):
    popSize = int(sum(iPop[:,1].toarray())[0])
    mdt = (popSize/cap)**A
    tau = tau_x * cap/popSize
    simTime = simTime + tau
    
    if iPop[iPop._shape[0]-1, 1] >= 1:
        iPop = sc.sparse.vstack([iPop, [iPop[iPop._shape[0]-1,0] + 1, 0]]).tocsr()
        
    iPop = rk4(iPop, tau, mdt, mut_prob, mut_effect)
    
    if iPop[0,1] < 1:
        iPop = iPop[1:iPop._shape[0]-1,:]
    
    return iPop, simTime
    